#define _XOPEN_SOURCE 500

#include "nbody.h"

#include <mpi.h>

#include <assert.h>
#include <ctype.h>
#include <fcntl.h>
#include <getopt.h>
#include <ieee754.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <time.h>
#include <unistd.h>

#ifdef USE_OMPSS2
#include <nanos6/debug.h>
#endif

#ifdef USE_OMP
#include <omp.h>
#endif

#ifdef USE_CUDA
#include <cuda_runtime.h>
#endif

static void *allocate(size_t size, int gpu);
static void deallocate(void *pointer, int gpu);

void nbody_generate_particles(const nbody_conf_t *conf, const nbody_file_t *file)
{
	char fname[1024];
	sprintf(fname, "%s.in", file->name);
	
	if (!access(fname, F_OK)) {
		return;
	}
	
	struct stat st = {0};
	if (stat("data", &st) == -1) {
		mkdir("data", 0755);
	}
	
	const int fd = open(fname, O_RDWR|O_CREAT|O_TRUNC, S_IRUSR|S_IRGRP|S_IROTH);
	if (fd < 0) {
		perror(NULL);
	}
	assert(fd >= 0);
	
	const int total_size = file->total_size;
	
	int err = ftruncate(fd, total_size);
	assert(!err);
	
	particles_block_t * const particles = mmap(NULL, total_size, PROT_WRITE|PROT_READ, MAP_SHARED, fd, 0);
	
	for (int b = 0; b < conf->num_blocks; b++) {
		nbody_particle_init(conf, particles+b);
	}
	
	err = munmap(particles, total_size);
	assert(!err);
	
	err = close(fd);
	assert(!err);
}

void nbody_check(const nbody_t *nbody)
{
	assert(nbody != NULL);
	
	// GPU ranks do not participate
	if (nbody->gpus_group) {
		return;
	}
	
	int rank;
	MPI_Comm_rank(nbody->COMM_GROUP, &rank);
	
	char fname[1024];
	sprintf(fname, "%s.ref", nbody->file.name);
	if (access(fname, F_OK) != 0) {
		if (!rank) fprintf(stderr, "Warning: %s file does not exist. Skipping the check...\n", fname);
		return;
	}
	
	const int fd = open(fname, O_RDONLY, 0);
	assert(fd >= 0);
	
	particles_block_t *reference = mmap(NULL, nbody->file.size, PROT_READ, MAP_SHARED, fd, nbody->file.offset);
	assert(reference != MAP_FAILED);
	
	int correct, correct_chunk;
	correct_chunk = nbody_compare_particles(nbody->local, reference, nbody->num_blocks);
	
	MPI_Reduce(&correct_chunk, &correct, 1, MPI_INT, MPI_LAND, 0, nbody->COMM_GROUP);
	
	if (!rank) {
		if (correct) {
			printf("Result validation: OK\n");
		} else {
			printf("Result validation: ERROR\n");
		}
	}
	
	int err = munmap(reference, nbody->file.size);
	assert(!err);
	
	err = close(fd);
	assert(!err);
}

void nbody_setup_file(const nbody_conf_t *conf, nbody_file_t *file, MPI_Comm comm)
{
	assert(conf != NULL);
	assert(file != NULL);
	
	int rank, rank_size;
	MPI_Comm_rank(comm, &rank);
	MPI_Comm_size(comm, &rank_size);
	
	file->total_size = conf->num_blocks * sizeof(particles_block_t);
	file->size = file->total_size / rank_size;
	file->offset = rank * file->size;
	
	sprintf(file->name, "%s-N2-%d-%d-%d", conf->name, conf->num_particles, BLOCK_SIZE, conf->timesteps);
}

particles_block_t *nbody_load_particles(const nbody_file_t *file)
{
	char fname[1024];
	sprintf(fname, "%s.in", file->name);
	
	struct stat st;
	stat(fname, &st);
	assert(st.st_size == file->total_size);
	
	const int fd = open(fname, O_RDONLY, 0);
	assert(fd >= 0);
	
	void * const file_particles = mmap(NULL, file->size, PROT_READ|PROT_WRITE, MAP_PRIVATE, fd, file->offset);
	assert(file_particles != MAP_FAILED);
	
	void *particles = allocate(file->size, 0);
	assert(particles != NULL);
	
	memcpy(particles, file_particles, file->size);
	
	int err = close(fd);
	assert(!err);
	
	err = munmap(file_particles, file->size);
	assert(!err);
	
	return particles;
}

static void nbody_distribute_work(const nbody_conf_t *conf, nbody_t *nbody)
{
	int world_rank, world_size;
	MPI_Comm_rank(MPI_COMM_WORLD, &world_rank);
	MPI_Comm_size(MPI_COMM_WORLD, &world_size);
	
	// Detect whether the node has GPUs
#ifdef USE_CUDA
	int device_count = 0;
	cudaError_t err = cudaGetDeviceCount(&device_count);
	if (err == cudaSuccess && device_count > 0) {
		nbody->gpus_group = 1;
	} else {
		nbody->gpus_group = 0;
	}
#else
	nbody->gpus_group = 1;
#endif
	
	int gpus_group[world_size];
	bzero(gpus_group, world_size * sizeof(int));
	
	gpus_group[world_rank] = nbody->gpus_group;
	MPI_Allgather(MPI_IN_PLACE, 1, MPI_INT, gpus_group, 1, MPI_INT, MPI_COMM_WORLD);
	
	int gpus_group_size = 0;
	for (int i = 0; i < world_size; ++i) {
		if (gpus_group[i]) {
			if (gpus_group_size == conf->max_gpu_ranks) {
				gpus_group[i] = 0;
			} else {
				++gpus_group_size;
			}
		}
	}
	assert(gpus_group_size <= conf->max_gpu_ranks);
	
	// Update my configuration
	nbody->gpus_group = gpus_group[world_rank];
	
	// Setup the group communicator
	MPI_Comm_split(MPI_COMM_WORLD, nbody->gpus_group, world_rank, &nbody->COMM_GROUP);
	
	int group_rank;
	MPI_Comm_rank(nbody->COMM_GROUP, &group_rank);
	
	int cpus_group_size = world_size - gpus_group_size;
	assert(cpus_group_size > 0);
	assert(gpus_group_size > 0);
	assert(conf->num_blocks % cpus_group_size == 0);
	assert(conf->num_blocks % gpus_group_size == 0);
	
	int cpus_group_ranks[cpus_group_size];
	int gpus_group_ranks[gpus_group_size];
	
	int cindex = 0;
	int gindex = 0;
	for (int i = 0; i < world_size; ++i) {
		if (gpus_group[i]) {
			gpus_group_ranks[gindex] = i;
			++gindex;
		} else {
			cpus_group_ranks[cindex] = i;
			++cindex;
		}
	}
	assert(cindex == cpus_group_size);
	assert(gindex == gpus_group_size);
	
	const int blocks_per_cpu_rank = conf->num_blocks / cpus_group_size;
	const int blocks_per_gpu_rank = conf->num_blocks / gpus_group_size;
	assert(blocks_per_cpu_rank > 0);
	assert(blocks_per_gpu_rank > 0);
	
	nbody->num_blocks = (nbody->gpus_group) ? blocks_per_gpu_rank : blocks_per_cpu_rank;
	nbody->num_particles = nbody->num_blocks * BLOCK_SIZE;
	
	const int *peers = (nbody->gpus_group) ? cpus_group_ranks : gpus_group_ranks;
	const int blocks_per_peer = (nbody->gpus_group) ? blocks_per_cpu_rank : blocks_per_gpu_rank;
	
	// Assign the corresponding peer rank for each block
	nbody->particles_block_peer = (int *) allocate(nbody->num_blocks * sizeof(int), 0);
	assert(nbody->particles_block_peer != NULL);
	
	for (int b = 0; b < nbody->num_blocks; ++b) {
		nbody->particles_block_peer[b] = peers[(b + nbody->num_blocks * group_rank) / blocks_per_peer];
	}
	
	nbody->cpus_group_size = cpus_group_size;
	nbody->gpus_group_size = gpus_group_size;
}

void nbody_setup(const nbody_conf_t *conf, nbody_t *nbody)
{
	assert(conf != NULL);
	assert(nbody != NULL);
	
	nbody_distribute_work(conf, nbody);
	
	const int gpu = nbody->gpus_group;
	const int particles_size = nbody->num_blocks * sizeof(particles_block_t);
	const int forces_size = nbody->num_blocks * sizeof(forces_block_t);
	
	if (!gpu) {
		// GPU ranks do not participate in accessing files
		int cpus_rank;
		MPI_Comm_rank(nbody->COMM_GROUP, &cpus_rank);
		
		nbody_setup_file(conf, &nbody->file, nbody->COMM_GROUP);
		
		if (cpus_rank == 0) nbody_generate_particles(conf, &nbody->file);
		MPI_Barrier(nbody->COMM_GROUP);
		
		nbody->local = nbody_load_particles(&nbody->file);
	} else {
		nbody->local = allocate(particles_size, gpu);
	}
	
	nbody->remote1 = (particles_block_t *) allocate(particles_size, gpu);
	nbody->remote2 = (particles_block_t *) allocate(particles_size, gpu);
	nbody->forces = (forces_block_t *) allocate(forces_size, gpu);
	
	assert(nbody->local != NULL);
	assert(nbody->remote1 != NULL);
	assert(nbody->remote2 != NULL);
	assert(nbody->forces != NULL);
	
	nbody->timesteps = conf->timesteps;
}

void nbody_free(nbody_t *nbody)
{
	const int gpu = nbody->gpus_group;
	deallocate(nbody->local, gpu);
	deallocate(nbody->remote1, gpu);
	deallocate(nbody->remote2, gpu);
	deallocate(nbody->forces, gpu);
	deallocate(nbody->particles_block_peer, 0);
}

void nbody_save_particles(const nbody_t *nbody)
{
	assert(nbody != NULL);
	
	// GPU ranks do not participate
	if (!nbody->gpus_group) {
		int rank, rank_size;
		MPI_Comm_rank(nbody->COMM_GROUP, &rank);
		MPI_Comm_size(nbody->COMM_GROUP, &rank_size);
		
		char fname[1024];
		sprintf(fname, "%s.out", nbody->file.name);
		const int mode = MPI_MODE_CREATE | MPI_MODE_WRONLY;
		
		MPI_File outfile;
		int err = MPI_File_open(nbody->COMM_GROUP, fname, mode, MPI_INFO_NULL, &outfile);
		assert(err == MPI_SUCCESS);
		
		err = MPI_File_set_view(outfile, nbody->file.offset, MPI_BYTE, MPI_BYTE, "native", MPI_INFO_NULL);
		assert(err == MPI_SUCCESS);
		
		err = MPI_File_write(outfile, nbody->local, nbody->file.size, MPI_BYTE, MPI_STATUS_IGNORE);
		assert(err == MPI_SUCCESS);
		
		err = MPI_File_close(&outfile);
		assert(err == MPI_SUCCESS);
	}
	
	MPI_Barrier(MPI_COMM_WORLD);
}

static void nbody_print_usage(int argc, char **argv)
{
	fprintf(stderr, "Usage: %s <-p particles> <-t timesteps> [OPTION]...\n", argv[0]);
	fprintf(stderr, "Parameters:\n");
	fprintf(stderr, "  -p, --particles=PARTICLES  the total number of particles (default: 16384)\n");
	fprintf(stderr, "  -t, --timesteps=TIMESTEPS  the number of timesteps (default: 10)\n\n");
	fprintf(stderr, "Optional parameters:\n");
	fprintf(stderr, "  -g, --max-gpu-ranks=NUM    the maximum number of GPU ranks (default: half of the total ranks)\n");
	fprintf(stderr, "  -c, --check                check the correctness of the result (default: disabled)\n");
	fprintf(stderr, "  -o, --output               save the computed particles to the default output file (default: disabled)\n");
	fprintf(stderr, "  -h, --help                 display this help and exit\n\n");
}

void nbody_get_conf(int argc, char **argv, nbody_conf_t *conf)
{
	assert(conf != NULL);
	
	int rank, rank_size;
	MPI_Comm_rank(MPI_COMM_WORLD, &rank);
	MPI_Comm_size(MPI_COMM_WORLD, &rank_size);
	
	if (rank == 0) {
		conf->domain_size_x = default_domain_size_x;
		conf->domain_size_y = default_domain_size_y;
		conf->domain_size_z = default_domain_size_z;
		conf->mass_maximum  = default_mass_maximum;
		conf->time_interval = default_time_interval;
		conf->seed          = default_seed;
		conf->num_particles = default_num_particles;
		conf->num_blocks    = conf->num_particles / BLOCK_SIZE;
		conf->timesteps     = default_timesteps;
		conf->save_result   = default_save_result;
		conf->check_result  = default_check_result;
		conf->max_gpu_ranks = rank_size / 2;
		strcpy(conf->name, default_name);
		
		static struct option long_options[] = {
			{"particles",     required_argument, 0, 'p'},
			{"timesteps",     required_argument, 0, 't'},
			{"max-gpu-ranks", required_argument, 0, 'g'},
			{"check",         no_argument,       0, 'c'},
			{"output",        no_argument,       0, 'o'},
			{"help",          no_argument,       0, 'h'},
			{0, 0, 0, 0}
		};
		
		int c;
		int index;
		while ((c = getopt_long(argc, argv, "hocp:t:g:", long_options, &index)) != -1) {
			switch (c) {
				case 'h':
					nbody_print_usage(argc, argv);
					exit(0);
				case 'o':
					conf->save_result = 1;
					break;
				case 'c':
					conf->check_result = 1;
					break;
				case 'p':
					conf->num_particles = atoi(optarg);
					break;
				case 't':
					conf->timesteps = atoi(optarg);
					break;
				case 'g':
					conf->max_gpu_ranks = atoi(optarg);
					break;
				case '?':
					exit(1);
				default:
					abort();
			}
		}
		
		if (conf->num_particles < 1 || conf->timesteps < 1 || conf->max_gpu_ranks < 1) {
			nbody_print_usage(argc, argv);
			exit(1);
		}
		
		conf->num_particles = ROUNDUP(conf->num_particles, MIN_PARTICLES);
		assert(conf->num_particles >= BLOCK_SIZE);
		
		conf->num_blocks = conf->num_particles / BLOCK_SIZE;
		assert(conf->num_blocks > 0);
	}
	
	MPI_Bcast(conf, sizeof(nbody_conf_t), MPI_BYTE, 0, MPI_COMM_WORLD);
	assert(conf->num_particles > 0);
	assert(conf->timesteps > 0);
	assert(conf->max_gpu_ranks > 0);
}

void nbody_particle_init(const nbody_conf_t *conf, particles_block_t *particles)
{
	for (int i = 0; i < BLOCK_SIZE; i++){
		particles->position_x[i] = conf->domain_size_x * ((float)random() / ((float)RAND_MAX + 1.0));
		particles->position_y[i] = conf->domain_size_y * ((float)random() / ((float)RAND_MAX + 1.0));
		particles->position_z[i] = conf->domain_size_z * ((float)random() / ((float)RAND_MAX + 1.0));
		particles->mass[i] = conf->mass_maximum * ((float)random() / ((float)RAND_MAX + 1.0));
		particles->weight[i] = gravitational_constant * particles->mass[i];
	}
}

int nbody_compare_particles(const particles_block_t *local, const particles_block_t *reference, int num_blocks)
{
	double error = 0.0;
	int count = 0;
	for (int i = 0; i < num_blocks; i++) {
		for (int e = 0; e < BLOCK_SIZE; e++) {
			if ((local[i].position_x[e] != reference[i].position_x[e]) ||
			    (local[i].position_y[e] != reference[i].position_y[e]) ||
			    (local[i].position_z[e] != reference[i].position_z[e])) {
					error += fabs(((local[i].position_x[e] - reference[i].position_x[e])*100.0) / reference[i].position_x[e]) +
					         fabs(((local[i].position_y[e] - reference[i].position_y[e])*100.0) / reference[i].position_y[e]) +
					         fabs(((local[i].position_z[e] - reference[i].position_z[e])*100.0) / reference[i].position_z[e]);
					count++;
			}
		}
	}
	
	double relative_error = (count != 0) ? error / (3.0 * count) : 0.0;
	if ((count * 100.0) / (num_blocks * BLOCK_SIZE) > 0.6 || relative_error > TOLERATED_ERROR) {
		return 0;
	}
	return 1;
}

double nbody_compute_throughput(int num_particles, int timesteps, double elapsed_time)
{
	double interactions_per_timestep = (double)(num_particles)*(double)(num_particles);
	return (((interactions_per_timestep * (double)timesteps) / elapsed_time) / 1000000.0);
}

void nbody_stats(const nbody_t *nbody, const nbody_conf_t *conf, double time)
{
	int rank, rank_size;
	MPI_Comm_rank(MPI_COMM_WORLD, &rank);
	MPI_Comm_size(MPI_COMM_WORLD, &rank_size);
	
	if (!rank) {
		int num_cpus = 1;
#if defined(USE_OMPSS2)
		num_cpus = nanos6_get_num_cpus();
#elif defined(USE_OMP)
		num_cpus = omp_get_max_threads();
#endif
		double throughput = nbody_compute_throughput(conf->num_particles, nbody->timesteps, time);
		
		printf("bigo, N2, cpu_ranks, %d, gpu_ranks, %d  threads, %d, timesteps, %d, total_particles, %d, block_size, %d, time, %.2f, throughput, %.2f\n",
			nbody->cpus_group_size, nbody->gpus_group_size, num_cpus, nbody->timesteps, conf->num_particles, BLOCK_SIZE, time, throughput
		);
	}
}

void *allocate(size_t size, int gpu)
{
	void *pointer = NULL;
#ifdef USE_CUDA
	if (gpu) {
		cudaError_t err = cudaMallocManaged(&pointer, size, cudaMemAttachGlobal);
		assert(err == cudaSuccess);
	} else
#endif
	{
		pointer = malloc(size);
	}
	assert(pointer != NULL);
	
	return pointer;
}

void deallocate(void *pointer, int gpu)
{
#ifdef USE_CUDA
	if (gpu) {
		cudaError_t err = cudaFree(pointer);
		assert(err == cudaSuccess);
	} else
#endif
	{
		free(pointer);
	}
}

double get_time()
{
	struct timespec ts;
	clock_gettime(CLOCK_MONOTONIC, &ts);
	return (double)(ts.tv_sec) + (double)ts.tv_nsec * 1.0e-9;
}
